{% load auth %}
<div{% include 'django/forms/widgets/attrs.html' %}>
  {% render_password_as_hash widget.value %}
  <p><a role="button" class="button" href="{{ password_url|default:"../password/" }}">{{ button_label }}</a></p>
</div>
