{% load i18n %}

<div id="{{ widget.attrs.id }}_div_map" class="dj_map_wrapper">
    <div id="{{ widget.attrs.id }}_map" class="dj_map"></div>
    {% if not disabled %}<span class="clear_features"><a href="">{% translate "Delete all Features" %}</a></span>{% endif %}
    {% if widget.attrs.display_raw %}<p>{% translate "Debugging window (serialized value)" %}</p>{% endif %}
    <textarea id="{{ widget.attrs.id }}" class="vSerializedField required" cols="150" rows="10" name="{{ widget.name }}"
              {% if not widget.attrs.display_raw %} hidden{% endif %}>{{ serialized }}</textarea>
    {% with script_id=widget.attrs.id|add:"_mapwidget_options" %}
    {{ widget.attrs|json_script:script_id }}
    {% endwith %}
</div>
