from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("lookuperror_c", "0002_c2"),
    ]

    operations = [
        migrations.CreateModel(
            name="C3",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        serialize=False,
                        verbose_name="ID",
                        primary_key=True,
                    ),
                ),
            ],
        ),
    ]
